Package: dynamite
Title: Bayesian Modeling and Causal Inference for Multivariate
        Longitudinal Data
Version: 1.6.3
Authors@R: c(
    person("Santtu", "Tikka", email = "santtuth@gmail.com", 
           role = c("aut", "cre"), comment = c(ORCID = "0000-0003-4039-4342")),
    person("Jouni", "Helske", email = "jouni.helske@iki.fi", role = "aut",
           comment = c(ORCID = "0000-0001-7130-793X")),
    person("Nicholas", "Clark", role = "rev"),
    person("Lucy", "D'Agostino McGowan", role = "rev")
  )
Description: Easy-to-use and efficient interface for 
  Bayesian inference of complex panel (time series) data using dynamic 
  multivariate panel models by Helske and Tikka (2024) 
  <doi:10.1016/j.alcr.2024.100617>. The package supports joint modeling of 
  multiple measurements per individual, time-varying and time-invariant 
  effects, and a wide range of discrete and continuous distributions. 
  Estimation of these dynamic multivariate panel models is carried out via 
  'Stan'. For an in-depth tutorial of the package, see 
  (Tikka and Helske, 2025) <doi:10.18637/jss.v115.i05>.
License: GPL (>= 3)
URL: https://docs.ropensci.org/dynamite/,
        https://github.com/ropensci/dynamite/
BugReports: https://github.com/ropensci/dynamite/issues/
Depends: R (>= 3.6.0)
Imports: checkmate, cli, data.table (>= 1.15.0), ggforce, glue,
        ggplot2, loo, patchwork, posterior, rlang, rstan, stats, tibble
        (>= 2.0.0), utils
Suggests: cmdstanr, covr, dplyr, knitr, mice, mockthat, quarto,
        testthat (>= 3.0.0), tidyr
VignetteBuilder: quarto
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
LazyData: true
LazyDataCompression: xz
Additional_repositories: https://stan-dev.r-universe.dev/
NeedsCompilation: no
Packaged: 2026-04-27 09:22:28 UTC; santikka
Author: Santtu Tikka [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4039-4342>),
  Jouni Helske [aut] (ORCID: <https://orcid.org/0000-0001-7130-793X>),
  Nicholas Clark [rev],
  Lucy D'Agostino McGowan [rev]
Maintainer: Santtu Tikka <santtuth@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-27 09:50:02 UTC
Built: R 4.6.0; ; 2026-04-27 21:39:41 UTC; unix
