# Generated by roxygen2: do not edit by hand

S3method("+",dynamiteformula)
S3method(as.data.frame,dynamitefit)
S3method(as.data.table,dynamitefit)
S3method(as_draws,dynamitefit)
S3method(as_draws_df,dynamitefit)
S3method(coef,dynamitefit)
S3method(confint,dynamitefit)
S3method(fitted,dynamitefit)
S3method(formula,dynamitefit)
S3method(get_algorithm,CmdStanMCMC)
S3method(get_algorithm,CmdStanMCMC_CSV)
S3method(get_algorithm,stanfit)
S3method(get_code,dynamitefit)
S3method(get_code,dynamiteformula)
S3method(get_data,dynamitefit)
S3method(get_data,dynamiteformula)
S3method(get_diagnostics,CmdStanMCMC)
S3method(get_diagnostics,CmdStanMCMC_CSV)
S3method(get_diagnostics,stanfit)
S3method(get_draws,CMdStanMCMC_CSV)
S3method(get_draws,CmdStanMCMC)
S3method(get_draws,stanfit)
S3method(get_elapsed_time,CmdStanMCMC)
S3method(get_elapsed_time,CmdStanMCMC_CSV)
S3method(get_elapsed_time,stanfit)
S3method(get_max_treedepth,CmdStanMCMC)
S3method(get_max_treedepth,CmdStanMCMC_CSV)
S3method(get_max_treedepth,stanfit)
S3method(get_model_code,CmdStanMCMC)
S3method(get_model_code,CmdStanMCMC_CSV)
S3method(get_model_code,stanfit)
S3method(get_nchains,CmdStanMCMC)
S3method(get_nchains,CmdStanMCMC_CSV)
S3method(get_nchains,stanfit)
S3method(get_ndraws,CmdStanMCMC)
S3method(get_ndraws,CmdStanMCMC_CSV)
S3method(get_ndraws,stanfit)
S3method(get_parameter_dims,dynamitefit)
S3method(get_parameter_dims,dynamiteformula)
S3method(get_parameter_names,dynamitefit)
S3method(get_parameter_types,dynamitefit)
S3method(get_pars_oi,CmdStanMCMC)
S3method(get_pars_oi,CmdStanMCMC_CSV)
S3method(get_pars_oi,stanfit)
S3method(get_priors,dynamitefit)
S3method(get_priors,dynamiteformula)
S3method(hmc_diagnostics,dynamitefit)
S3method(lfo,dynamitefit)
S3method(loo,dynamitefit)
S3method(mcmc_diagnostics,dynamitefit)
S3method(ndraws,dynamitefit)
S3method(nobs,dynamitefit)
S3method(plot,dynamitefit)
S3method(plot,dynamiteformula)
S3method(plot,lfo)
S3method(predict,dynamitefit)
S3method(print,dynamitefit)
S3method(print,dynamiteformula)
S3method(print,lfo)
S3method(summary,dynamitefit)
S3method(update,dynamitefit)
export(as.data.table)
export(as_draws)
export(as_draws_df)
export(aux)
export(dynamice)
export(dynamite)
export(dynamiteformula)
export(get_algorithm)
export(get_code)
export(get_data)
export(get_diagnostics)
export(get_draws)
export(get_elapsed_time)
export(get_max_treedepth)
export(get_model_code)
export(get_nchains)
export(get_ndraws)
export(get_parameter_dims)
export(get_parameter_names)
export(get_parameter_types)
export(get_pars_oi)
export(get_priors)
export(hmc_diagnostics)
export(lags)
export(lfactor)
export(lfo)
export(loo)
export(mcmc_diagnostics)
export(mice.impute.lag)
export(mice.impute.lead)
export(ndraws)
export(obs)
export(plot_betas)
export(plot_deltas)
export(plot_lambdas)
export(plot_nus)
export(plot_psis)
export(random_spec)
export(splines)
importFrom(data.table,":=")
importFrom(data.table,as.data.table)
importFrom(loo,loo)
importFrom(patchwork,wrap_plots)
importFrom(posterior,as_draws)
importFrom(posterior,as_draws_df)
importFrom(posterior,ndraws)
importFrom(rlang,.data)
importFrom(stats,as.formula)
importFrom(stats,coef)
importFrom(stats,fitted)
importFrom(stats,formula)
importFrom(stats,model.matrix)
importFrom(stats,model.matrix.lm)
importFrom(stats,na.exclude)
importFrom(stats,na.pass)
importFrom(stats,nobs)
importFrom(stats,plogis)
importFrom(stats,qlogis)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,reformulate)
importFrom(stats,rnbinom)
importFrom(stats,rnorm)
importFrom(stats,rpois)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,terms)
importFrom(stats,update)
